from lxml import etree, builder

import sword2
from .exceptions import (
    MissingMandatoryMetadataException,
    NotSoftwareRecordException,
    RecordHasNoFilesException,
)

CodeMeta = builder.ElementMaker(namespace="https://doi.org/10.5063/SCHEMA/CODEMETA-2.0")


class SWHMetadata:
    namespaces = {
        "codemeta": "https://doi.org/10.5063/SCHEMA/CODEMETA-2.0",
        "swh": "https://www.softwareheritage.org/schema/2018/deposit",
    }

    def __init__(self, extension):
        self.extension = extension

    def __call__(self, data: dict) -> sword2.Entry:
        if not self.is_software_record_data(data):
            raise NotSoftwareRecordException

        if data.get("files", {}).get("enabled") is False:
            raise RecordHasNoFilesException(
                "Depositing to Software Heritage requires a non-metadata-only record."
            )

        entry = sword2.Entry()
        for prefix in self.namespaces:
            entry.register_namespace(prefix, self.namespaces[prefix])

        self.add_atom_metadata(entry, data)
        self.add_codemeta_metadata(entry, data)
        self.add_swh_metadata(entry, data)

        return entry

    def add_atom_metadata(self, entry: sword2.Entry, data: dict) -> None:
        try:
            entry.add_field("title", data["metadata"]["title"])
        except KeyError as e:
            raise MissingMandatoryMetadataException("A title is required.") from e

        if not data["metadata"].get("creators"):
            raise MissingMandatoryMetadataException("At least one creator is required.")

        for creator in data["metadata"].get("creators"):
            # TODO: The schema says identifiers are a list of dicts
            orcid_uri = (
                ("https://orcid.org/" + creator["identifiers"]["orcid"])
                if creator.get("identifiers", {}).get("orcid")
                else None
            )
            # TODO: The schema says this should be creator['person_or_org']['name']
            entry.add_author(name=creator["name"], uri=orcid_uri)

    def add_codemeta_metadata(self, entry: sword2.Entry, data: dict) -> None:
        for date in data["metadata"].get("dates", []):
            if not date.get("date"):
                continue
            date_type, date_value = date.get("type"), date["date"]
            if date_type == "created":
                entry.add_field("codemeta_dateCreated", date_value)
            if date_type == "updated":
                entry.add_field("codemeta_datePublished", date_value)

        for rights in data["metadata"].get("rights", []):
            entry.entry.append(
                CodeMeta.license(
                    CodeMeta.name(rights["rights"]),
                    CodeMeta.url(rights["url"]),
                )
            )

    def add_swh_metadata(self, entry: sword2.Entry, data: dict) -> None:
        pass

    def is_software_record_data(self, data: dict) -> bool:
        try:
            return data["metadata"]["resource_type"]["type"] == "software"
        except KeyError:
            return False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN
# Copyright (C) 2020 Cottage Labs LLP.
#
# invenio-swh is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Version information for invenio-swh.

This file is imported by ``invenio_swh.__init__``,
and parsed by ``setup.py``.
"""

__version__ = '0.1.dev2'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN
# Copyright (C) 2020 Cottage Labs LLP.
#
# invenio-swh is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Support for onward deposit of software artifacts to Software Heritage"""


INVENIO_SWH_BASE_TEMPLATE = 'invenio_swh/base.html'
"""Default base template for the demo page."""

INVENIO_SWH_SERVICES = {
    'swh': {
        'NAME': "Software Heritage",
        'SERVICE_DOCUMENT': 'https://deposit.softwareheritage.org/1/servicedocument/',
        'USERNAME': None,
        'PASSWORD': None,
    }
}



class InvenioSWHException(Exception):
    annotate_record = False
    """Whether to annotate a record when this exception is raised.

    Errors may occur during processing which should be fed back to the user. To support
    this, code that performs actions should set `record['ext']['swh']['error']` to the
    string value of the exception if `annotate_record` is True. If no exceptions occur
    during an operation, this key should be cleared.
    """


class MissingMandatoryMetadataException(InvenioSWHException, ValueError):
    annotate_record = True


class NotSoftwareRecordException(InvenioSWHException, ValueError):
    pass


class RecordHasNoFilesException(InvenioSWHException):
    annotate_record = True



import logging
import typing
from flask import current_app
from invenio_drafts_resources.records import Draft
from invenio_records import Record
from lxml import etree

from invenio_rdm_records.records import RDMDraft
from invenio_records_resources.services.records.components import ServiceComponent
from invenio_swh import InvenioSWH
from invenio_swh.exceptions import (
    InvenioSWHException,
    MissingMandatoryMetadataException,
)
from sword2 import Deposit_Receipt

logger = logging.getLogger(__name__)

logging.basicConfig(level=logging.DEBUG)


class InvenioSWHComponent(ServiceComponent):
    """A service component providing SWH integration with records."""

    user_ext_key = "swh"
    internal_ext_key = "swh-internal"

    def __init__(self, service, *, extension_name="invenio-swh"):
        super().__init__(service)
        self.extension_name = extension_name

    def create(self, identity, *, data, record):
        logger.debug("Record create (in_progress=%s)", isinstance(record, Draft))
        self.sync_to_swh(data, record, in_progress=isinstance(record, Draft))

    def update(self, identity, *, data, record):
        logger.debug("Record update (in_progress=%s)", isinstance(record, Draft))
        self.sync_to_swh(data, record, in_progress=False)

    def publish(self, *, draft, record):
        logger.debug("Record publish")
        internal_data = self.get_extension_data(record, self.internal_ext_key)
        if internal_data.get("se-iri"):
            client = self.extension.sword_client
            client.complete_deposit(se_iri=internal_data["se-iri"])

    def read(self, identity, *, record):
        # Hide our internal metadata from the search index and the user
        # self.set_extension_data(record, self.internal_ext_key, None)
        pass

    def update_draft(self, identity, *, data, record: RDMDraft):
        logger.debug("Record update draft")
        self.sync_to_swh(data, record, in_progress=True)

    def sync_to_swh(self, data: dict, record: Record, in_progress: bool):
        user_data = self.get_extension_data(record, self.user_ext_key)
        internal_data = self.get_extension_data(record, self.internal_ext_key)

        # Clear any error information
        user_data.pop("error", None)

        try:
            metadata_entry = self.extension.metadata(data)
            logger.info(
                "Extracted metadata for deposit: %s",
                etree.tounicode(metadata_entry.entry),
            )
        except InvenioSWHException as e:
            if e.annotate_record:
                user_data["error"] = str(e)
            logger.debug("Not extracting metadata for SWH deposit", exc_info=e)
            metadata_entry = None

        client = self.extension.sword_client

        result: typing.Optional[Deposit_Receipt]

        if internal_data.get("edit-iri") and metadata_entry:
            result = client.update(
                edit_iri=internal_data["edit-iri"],
                in_progress=in_progress,
                metadata_entry=self.extension.metadata(data),
            )
        elif metadata_entry:
            result = client.create(
                col_iri=self.extension.collection_iri,
                in_progress=in_progress,
                metadata_entry=self.extension.metadata(data),
            )
        else:
            result = None

        if result:
            internal_data.update(
                {
                    "edit-iri": result.edit,
                    "edit-media-iri": result.edit_media,
                    "se-iri": result.se_iri,
                }
            )

        self.set_extension_data(record, self.internal_ext_key, user_data)
        self.set_extension_data(record, self.internal_ext_key, internal_data)

    def get_extension_data(self, record: Record, key: str) -> dict:
        return record.get("ext", {}).get(key, {})

    def set_extension_data(self, record: Record, key: str, extension_data) -> dict:
        if extension_data:
            if "ext" not in record:
                record["ext"] = {}
            record["ext"][key] = extension_data
        elif "ext" in record:
            record["ext"].pop(str, None)
            if not record["ext"]:
                del record["ext"]

    @property
    def extension(self) -> InvenioSWH:
        """Returns the associated invenio-swh extension for this component"""
        return current_app.extensions[self.extension_name]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN
# Copyright (C) 2020 Cottage Labs LLP.
#
# invenio-swh is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Support for onward deposit of software artifacts to Software Heritage"""

# TODO: This is an example file. Remove it if you do not need it, including
# the templates and static folders as well as the test case.

from flask import Blueprint, render_template
from flask_babelex import gettext as _

blueprint = Blueprint(
    'invenio_swh',
    __name__,
    template_folder='templates',
    static_folder='static',
)


@blueprint.route("/")
def index():
    """Render a basic view."""
    return render_template(
        "invenio_swh/index.html",
        module_name=_('invenio-swh'))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN
# Copyright (C) 2020 Cottage Labs LLP.
#
# invenio-swh is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Support for onward deposit of software artifacts to Software Heritage"""

from .ext import InvenioSWH
from .version import __version__

__all__ = ('__version__', 'InvenioSWH')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN
# Copyright (C) 2020 Cottage Labs LLP.
#
# invenio-swh is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Support for onward deposit of software artifacts to Software Heritage"""

import os

from setuptools import find_packages, setup

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

tests_require = [
    "pytest-invenio>=1.4.0",
    "invenio-db[postgresql,versioning]",
    "invenio-search[elasticsearch7]",
    "pytest-invenio>=1.4.0",
    "xmldiff",
]

extras_require = {
    'docs': [
        'Sphinx>=3',
    ],
    'tests': tests_require,
}

extras_require['all'] = []
for reqs in extras_require.values():
    extras_require['all'].extend(reqs)

setup_requires = [
    'Babel>=2.8',
]

install_requires = [
    'invenio-i18n>=1.2.0',
    "invenio-rdm-records",
    "lxml",
    "sword2",
]

packages = find_packages()


# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('invenio_swh', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name='invenio-swh',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    keywords='invenio TODO',
    license='MIT',
    author='Cottage Labs LLP',
    author_email='info@inveniosoftware.org',
    url='https://github.com/inveniosoftware/invenio-swh',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    entry_points={
        'invenio_base.apps': [
            'invenio_swh = invenio_swh:InvenioSWH',
        ],
        'invenio_base.blueprints': [
            'invenio_swh = invenio_swh.views:blueprint',
        ],
        'invenio_i18n.translations': [
            'messages = invenio_swh',
        ],
        # TODO: Edit these entry points to fit your needs.
        # 'invenio_access.actions': [],
        # 'invenio_admin.actions': [],
        # 'invenio_assets.bundles': [],
        # 'invenio_base.api_apps': [],
        # 'invenio_base.api_blueprints': [],
        # 'invenio_base.blueprints': [],
        # 'invenio_celery.tasks': [],
        # 'invenio_db.models': [],
        # 'invenio_pidstore.minters': [],
        # 'invenio_records.jsonresolver': [],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Development Status :: 1 - Planning',
    ],
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN
# Copyright (C) 2020 Cottage Labs LLP.
#
# invenio-swh is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Support for onward deposit of software artifacts to Software Heritage"""

import typing
from flask import current_app

import sword2
from . import config
from .metadata import SWHMetadata


class InvenioSWH(object):
    """invenio-swh extension."""

    extension_name = "invenio-swh"

    url_config_key = "INVENIO_SWH_SERVICE_DOCUMENT"
    collection_name_config_key = "INVENIO_SWH_COLLECTION_NAME"
    collection_iri_config_key = "INVENIO_SWH_COLLECTION_IRI"
    username_config_key = "INVENIO_SWH_USERNAME"
    password_config_key = "INVENIO_SWH_PASSWORD"

    metadata_cls: typing.Type[SWHMetadata] = SWHMetadata

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        app.extensions[self.extension_name] = self

    def init_config(self, app):
        """Initialize configuration."""
        # Use theme's base template if theme is installed
        if "BASE_TEMPLATE" in app.config:
            app.config.setdefault(
                "INVENIO_SWH_BASE_TEMPLATE",
                app.config["BASE_TEMPLATE"],
            )
        for k in dir(config):
            if k.startswith("INVENIO_SWH_"):
                app.config.setdefault(k, getattr(config, k))

    @property
    def sword_client(self) -> sword2.Connection:
        if self.is_configured:
            return sword2.Connection(
                service_document_iri=current_app.config[self.url_config_key],
                user_name=current_app.config[self.username_config_key],
                user_pass=current_app.config[self.password_config_key],
            )

    @property
    def collection_name(self) -> str:
        return current_app.config[self.collection_name_config_key]

    @property
    def collection_iri(self) -> str:
        return current_app.config[self.collection_iri_config_key]

    @property
    def is_configured(self) -> bool:
        return bool(
            current_app.config.get(self.url_config_key)
            and current_app.config.get(self.collection_name_config_key)
            and current_app.config.get(self.username_config_key)
            and current_app.config.get(self.password_config_key)
        )

    @property
    def metadata(self) -> SWHMetadata:
        return self.metadata_cls(self)
